import os
import torch
from typing import Optional, List
from abc import ABC,abstractmethod
from dataclasses import dataclass, field
from torch.utils.data import DataLoader, Dataset, RandomSampler


@dataclass
class DataModuleConfig:
    """
    Data configuration class for defining data-related configuration parameters.

    Args:
        train_data_path (str): File path to the training data.
        val_data_path (str): File path to the validation data.
        test_data_path (str): File path to the test data.
        do_setup (bool): Whether prepare datasets from data_paths
        cache_dir (Optional[str]): Path to cache the datasets
        train_batch_size (Optional[int]): Batch size for training data. Defaults to 32.
        train_limit_batches (Optional[float]): Batch limit ratio for training data. Defaults to 1.0.
        val_batch_size (Optional[int]): Batch size for validation data. Defaults to 32.
        val_limit_batches (Optional[float]): Batch limit ratio for validation data. Defaults to 1.0.
        test_batch_size (Optional[int]): Batch size for test data. Defaults to 32.
        test_limit_batches (Optional[float]): Batch limit ratio for test data. Defaults to 1.0.
    """
    train_data_path: str = field(default=None, metadata={"description": "File path to the training data"})
    val_data_path: str = field(default=None, metadata={"description": "File path to the validation data"})
    test_data_path: str = field(default=None, metadata={"description": "File path to the test data"})
    do_setup: bool = field(default=True, metadata={"description": "Whether prepare datasets from data_paths"})
    cache_dir: str = field(default=None, metadata={"description": "Path to cache the datasets"})
    
    train_batch_size: int = field(default=32, metadata={"description": "Batch size for training data"})
    train_limit_batches: float = field(default=1.0, metadata={"description": "Batch limit ratio for training data"})
    val_batch_size: int = field(default=32, metadata={"description": "Batch size for validation data"})
    val_limit_batches: float = field(default=1.0, metadata={"description": "Batch limit ratio for validation data"})
    test_batch_size: int = field(default=32, metadata={"description": "Batch size for test data"})
    test_limit_batches: float = field(default=1.0, metadata={"description": "Batch limit ratio for test data"})


class DataModule(ABC):
    def __init__(self, 
        config: DataModuleConfig):
        """
        Abstract base class for data modules.

        Args:
            config (DataModuleConfig): Data configuration object.
        """
        self.config = config
        if self.config.cache_dir and not self.config.do_setup:
            self.train_dataset = torch.load(os.path.join(self.config.cache_dir, "train_dataset.pt"))
            self.val_dataset = torch.load(os.path.join(self.config.cache_dir, "val_dataset.pt"))
            self.test_dataset = torch.load(os.path.join(self.config.cache_dir, "test_dataset.pt"))
        else:
            self.setup()
            if self.config.cache_dir:
                torch.save(self.train_dataset,os.path.join(self.config.cache_dir, "train_dataset.pt"))
                torch.save(self.val_dataset, os.path.join(self.config.cache_dir, "val_dataset.pt"))
                torch.save(self.test_dataset, os.path.join(self.config.cache_dir, "test_dataset.pt"))

    @abstractmethod
    def setup(self):
        """
        Prepare datasets from data_paths
        """
        pass

    @abstractmethod
    def train_dataloader(self):
        """
        Returns a data loader for training data.

        Returns:
            DataLoader: Data loader for training data.
        """
        full_size = len(self.train_dataset)
        real_size = int(full_size * self.config.train_limit_batches)
        train_dataset, _ = torch.utils.data.random_split(self.train_dataset, [real_size, full_size-real_size])
        train_dataloader = DataLoader(dataset = train_dataset, 
                                shuffle = True,
                                batch_size = self.config.train_batch_size)
        return train_dataloader
    
    @abstractmethod
    def val_dataloader(self):
        """
        Returns a data loader for validation data.

        Returns:
            DataLoader: Data loader for validation data.
        """
        full_size = len(self.val_dataset)
        real_size = int(full_size * self.config.val_limit_batches)
        val_dataset, _ = torch.utils.data.random_split(self.val_dataset, [real_size, full_size-real_size])
        val_dataloader = DataLoader(val_dataset, 
                                shuffle = False,
                                batch_size = self.config.val_batch_size)
        return val_dataloader

    @abstractmethod
    def test_dataloader(self):
        """
        Returns a data loader for test data.

        Returns:
            DataLoader: Data loader for test data.
        """
        full_size = len(self.test_dataset)
        real_size = int(full_size * self.config.test_limit_batches)
        test_dataset, _ = torch.utils.data.random_split(self.test_dataset, [real_size, full_size-real_size])
        test_dataloader = DataLoader(test_dataset, 
                                shuffle = False,
                                batch_size = self.config.test_batch_size)
        return test_dataloader


import os
import yaml
import torch
import random
import argparse
import numpy as np
from loguru import logger
from attrdict import AttrDict

__all__ = [
    "parse_args",
    "seed_everything",
]

def parse_args():
    """
    Parse command-line arguments and load the configuration file.

    Returns:
        config (AttrDict): Configuration loaded from the file and command-line arguments.
    """
    parser = argparse.ArgumentParser()

    parser.add_argument('--config_file', type=str, default='config.yaml',
                        help='Path to the configuration file')
    # parser.add_argument('--do_test', action='store_true', default=False,
    #                 help='Train or Test, True for test, False for train')
    
    args = parser.parse_args()
    config = AttrDict(
        yaml.load(open(args.config_file, 'r', encoding='utf-8'), Loader=yaml.FullLoader))
    # Update the configuration with command-line arguments
    for k, v in vars(args).items():
        setattr(config, k, v)

    return config


def seed_everything(seed: int) -> int:
    """
    Function that sets seed for pseudo-random number generators in:
    pytorch, numpy, python.random

    Args:
        seed: the integer value seed for global random state
    """
    print(f"Global seed set to {seed}")
    os.environ["PL_GLOBAL_SEED"] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    return seed


import os
import json
import datetime
from abc import ABC,abstractmethod
from .trainer import TrainingArguments
from .utils import seed_everything, parse_args, logger


class Runner(ABC):
    """
    Abstract base class for a runner.

    Subclasses of `Runner` should implement the `setup` method to initialize the model, data module and trainer class.
    The `run` method can be called to start the training or testing process based on the config file or command line arguments.

    Example:
        class MyRunner(Runner):
            def setup(self):
                model = BERT(config=BertConfig(**self.args.bert_config))
                data_module = KaggleDataModule(DataModuleConfig(**self.args.data_module_config))
                trainer_class = BertTrainer
                return model, data_module, trainer_class

        runner = MyRunner()
        runner.run()
    """
    def __init__(self) -> None:
        super().__init__()
        self.args = parse_args()
        logger.add(f"logs/{datetime.datetime.now().date()}/{self.args.log_file}", format="<green>{time:YYYY/MM/DD HH:mm}</green> |<level>{level}</level>| <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>") # Set path for logs
        logger.add(f"logs/{datetime.datetime.now().date()}/summary_{self.args.log_file}", format="<green>{time:YYYY/MM/DD HH:mm}</green> |<level>{level}</level>| <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>", level="SUCCESS")
        # logger.info("Configuration:" + json.dumps(self.args, indent=4))
        logger.success(f"CONFIG:{self.args.config_file}")
        # logger.info(f"python {os.path.basename(__file__)}!!!")
        seed_everything(self.args.random_seed)

    @abstractmethod
    def setup(self):
        """Initialize the model & data_module & trainer_class

        Returns:
            model: nn.Module
                An instance of nn.Module representing the initialized model.
            data_module: DataModule
                An instance of DataModule representing the initialized data module.
            trainer_class: Trainer
                A subclass of Trainer representing the trainer class available for training.
        """
        pass


    @logger.catch
    def run(self):
        """Run the training or testing process.

        This method initializes the model and data module, creates a trainer instance,
        and starts the training or testing based on the command line arguments.

        """
        start_time = datetime.datetime.now()
        
        # Initialize model & data_module
        model, data_module, trainer_class = self.setup()
            
        # Initialize trainer
        trainer = trainer_class(
            args=TrainingArguments(**self.args.training_args),
            model=model,
            data_module=data_module
        )

        if not self.args.do_test:    # Start training
            trainer.fit()
        else:                   # Start testing
            trainer.test(checkpoint=self.args.training_args.checkpoint_save_path)
                    
        end_time = datetime.datetime.now()
        training_duration = end_time - start_time
        training_minutes = training_duration.total_seconds() / 60
        logger.info(f"RUN TIME: {training_minutes} minutes. \n\n")




from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name='alegant',
    version='1.0.5',
    packages=find_packages(exclude=["dist.*", "dist", "*tests*", "*script*", "*cache*", "data"]),
    url='https://github.com/Hugo-Zhu/alegant',
    license='MIT',
    author='Haohao Zhu',
    author_email='zhuhh17@qq.com',
    description='Alegant: a elegant training framework',
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
                    'attrdict',
                    'loguru',
                    'numpy', 
                    'pyyaml',
                    'six',     # required by tensorboard
                    'tensorboard',
                    'torch',
                    'tqdm',
                    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)

